@@ -159,7 +159,7 @@ void PyDUChainTest::initShell()
 
 ReferencedTopDUContext PyDUChainTest::parse(const QString& code)
 {
-    TestFile* testfile = new TestFile(code + "\n", "py", 0, testDir.absolutePath().append("/"));
+    TestFile* testfile = new TestFile(code + "\n", "py", nullptr, testDir.absolutePath().append("/"));
     createdFiles << testfile;
 
     testfile->parse((TopDUContext::Features) (TopDUContext::ForceUpdate | TopDUContext::AST) );
@@ -171,7 +171,7 @@ ReferencedTopDUContext PyDUChainTest::parse(const QString& code)
         return testfile->topContext();
     }
     else Q_ASSERT(false && "Timed out waiting for parser results, aborting all tests");
-    return 0;
+    return nullptr;
 }
 
 PyDUChainTest::~PyDUChainTest()
@@ -278,6 +278,7 @@ void PyDUChainTest::testCrashes_data() {
     QTest::newRow("func_call") << "a = xxx.func(yyy.zzz)";
     QTest::newRow("comprehension_attrib") << "a = [foo for foo in bar].baz";
     QTest::newRow("comprehension_attrib2") << "a = [foo.bar for foo in bar]";
+    QTest::newRow("lambda_cmpr_defarg") << "a = lambda foo=[b for b in (1, 2, 3)]: foo";
     QTest::newRow("attrib") << "(sep or ' ').join(xxxx.capitalize() for xxxx in ssss.split(sep))";
     QTest::newRow("attrib2") << "(sep or ' ').join(x.capitalize() for x in s.split(sep))";
     QTest::newRow("attrib3") << "known_threads = {line.strip()}";
@@ -441,6 +442,7 @@ void PyDUChainTest::testCrashes_data() {
     QTest::newRow("comprehension_in_fstring") <<
         "def crash(): return f'expr={ {x: y for x, y in [(1, 2), ]}}'";
 #endif
+    QTest::newRow("comprehension_in_lambda") << "lambda foo: [bar for bar in foo]";
 }
 
 void PyDUChainTest::testClassVariables()
@@ -480,6 +482,7 @@ void PyDUChainTest::testWarnNewNotCls_data()
     QTest::newRow("check_first_arg_class_self_0") << "class c():\n def test(self, masik):\n  pass" << 0;
 }
 
+// this is actually for both binary and boolean operators
 void PyDUChainTest::testBinaryOperatorsUnsure()
 {
     QFETCH(QString, code);
@@ -504,6 +507,11 @@ void PyDUChainTest::testBinaryOperatorsUnsure_data()
     QTest::newRow("check_unsure_type_1") << "class c():\n def __mul__(self, other):\n  return int();\nx = c();\nx = 3;\ny = 3;\ncheckme = x * y;" << "int";
     QTest::newRow("check_unsure_type_2") << "class c():\n pass;\nx = c();\nx = 3;\ny = 3;\ncheckme = x * y;" << "int";
     QTest::newRow("check_unsure_type_3") << "class c():\n pass;\nclass d():\n pass;\nx = c();\nx = d();\ny = 3;\ncheckme = x * y;" << "int";
+
+    QTest::newRow("check_unsure_type_4") << "checkme = True or False" << "bool";
+    QTest::newRow("check_unsure_type_5") << "a = 'foo'; checkme = a or 'bar';" << "str";
+    QTest::newRow("check_unsure_type_6") << "class A(): pass\nclass B(): pass;\ncheckme = A() or B() or 42;" << "unsure (A, B, int)";
+
 }
 
 
@@ -817,7 +825,6 @@ void PyDUChainTest::testTypes()
     visitor->searchingForType = expectedType;
     visitor->visitCode(m_ast.data());
     QEXPECT_FAIL("tuple_func", "no suitable docstring hint", Continue);
-    QEXPECT_FAIL("tuple_slice", "not implemented", Continue);
     QEXPECT_FAIL("tuple_add", "not implemented", Continue);
     QEXPECT_FAIL("tuple_mul", "not implemented", Continue);
     QEXPECT_FAIL("return_builtin_iterator", "fake builtin iter()", Continue);
@@ -921,7 +928,12 @@ void PyDUChainTest::testTypes_data()
     QTest::newRow("diff_local_classattr3") << "attr=3.5\nclass c(): attr = 1\ninst=c()\ncheckme = c.attr" << "int";
 //     QTest::newRow("class_method_self") << "class c:\n def func(checkme, arg, arg2):\n  pass\n" << "c";
 //    QTest::newRow("funccall_dict") << "def foo(): return foo; checkme = foo();" << (uint) IntegralType::TypeFunction;
-    
+
+    // With only one subbed value we get a FormattedValue node
+    QTest::newRow("fstring_formattedvalue") << "name = 'Jim'; checkme = f'{name}'" << "str";
+    // Otherwise a JoinedString, with FormattedValues as children.
+    QTest::newRow("fstring_joinedstring") << "name = 'Jim'; checkme = f'Hello, {name}! Your name is {name}.'" << "str";
+
     QTest::newRow("tuple_simple") << "mytuple = 3, 5.5\ncheckme, foobar = mytuple" << "int";
     QTest::newRow("tuple_simple2") << "mytuple = 3, 5.5\nfoobar, checkme = mytuple" << "float";
     QTest::newRow("tuple_simple3") << "mytuple = 3, 5.5, \"str\", 3, \"str\"\na, b, c, d, checkme = mytuple" << "str";
@@ -978,6 +990,23 @@ void PyDUChainTest::testTypes_data()
     QTest::newRow("tuple_indexaccess_neg3") << "t = 3, 4.5; checkme = t[-3]" << "unsure (int, float)";
 
     QTest::newRow("tuple_slice") << "t = 3, 'q', 4.5; checkme = t[-3: 2]" << "tuple of (int, str)";
+    QTest::newRow("tuple_slice_normal")   << "t = 1, 2.3, 'a', {}; checkme = t[1:3]"    << "tuple of (float, str)";
+    QTest::newRow("tuple_slice_defstart") << "t = 1, 2.3, 'a', {}; checkme = t[:3]"     << "tuple of (int, float, str)";
+    QTest::newRow("tuple_slice_defstop")  << "t = 1, 2.3, 'a', {}; checkme = t[1:]"     << "tuple of (float, str, dict)";
+    QTest::newRow("tuple_slice_defboth")  << "t = 1, 2.3, 'a', {}; checkme = t[:]"      << "tuple of (int, float, str, dict)";
+    QTest::newRow("tuple_slice_step")     << "t = 1, 2.3, 'a', {}; checkme = t[0:3:2]"  << "tuple of (int, str)";
+    QTest::newRow("tuple_slice_reverse")  << "t = 1, 2.3, 'a', {}; checkme = t[3:1:-1]" << "tuple of (dict, str)";
+    QTest::newRow("tuple_slice_revstart") << "t = 1, 2.3, 'a', {}; checkme = t[:1:-1]"  << "tuple of (dict, str)";
+    QTest::newRow("tuple_slice_revstop")  << "t = 1, 2.3, 'a', {}; checkme = t[2::-1]"  << "tuple of (str, float, int)";
+    QTest::newRow("tuple_slice_revstop")  << "t = 1, 2.3, 'a', {}; checkme = t[::-1]"   << "tuple of (dict, str, float, int)";
+    QTest::newRow("tuple_slice_no_elems") << "t = 1, 2.3, 'a', {}; checkme = t[1:1]"    << "tuple of ()";
+    // TODO unsure-tuples.
+    QTest::newRow("tuple_slice_not_literal") << "n = 2; t = 1, 2.3, 'a', {}; checkme = t[0:n]" << "tuple of ()";
+    // These are allowed, for whatever reason.
+    QTest::newRow("tuple_slice_past_range") << "t = 1, 2.3; checkme = t[-999999999:8888888888]" << "tuple of (int, float)";
+    QTest::newRow("tuple_slice_wrong_direction") << "t = 1, 2.3, 'a'; checkme = t[0:3:-1]" << "tuple of ()";
+    // This isn't.
+    QTest::newRow("tuple_slice_zero_step") << "t = 1, 2.3; checkme = t[::0]" << "tuple of ()";
 
     QTest::newRow("tuple_add") << "t, u = (3,), ('q', 4.5); checkme = t + u" << "tuple of (int, str, float)";
     QTest::newRow("tuple_mul") << "t = 3, 4.5; checkme = t * 2" << "tuple of (int, float, int, float)";
@@ -1169,6 +1198,23 @@ void PyDUChainTest::testTypes_data()
     QTest::newRow("vararg_constructor") << "class myclass():\n"
                                            "  def __init__(self, *arg): self.prop = arg[0]\n"
                                            "obj = myclass(3, 5); checkme = obj.prop" << "int";
+
+    QTest::newRow("declaration_order_var") << "aaa = 2\n"
+                                              "checkme = aaa" << "int";
+    QTest::newRow("declaration_order_var2") << "checkme = aaa\n"
+                                               "aaa = 2" << "mixed";
+    QTest::newRow("declaration_order_func_defarg") << "aaa = 2\n"
+                                                      "def foo(x=aaa): return x\n"
+                                                      "checkme = foo()" << "int";
+    QTest::newRow("declaration_order_func_defarg2") << "def foo(x=aaa): return x\n"
+                                                       "aaa = 2\n"
+                                                       "checkme = foo()" << "mixed";
+    QTest::newRow("declaration_order_func_body") << "aaa = 2\n"
+                                                    "def foo(): return aaa\n"
+                                                   "checkme = foo()" << "int";
+    QTest::newRow("declaration_order_func_body2") << "def foo(): return aaa\n"
+                                                     "aaa = 2\n"
+                                                     "checkme = foo()" << "int";
     QTest::newRow("global_variable") << "a = 3\n"
                                         "def f1():\n"
                                         "  global a\n"
@@ -1189,11 +1235,49 @@ void PyDUChainTest::testTypes_data()
                                         "  return a\n"
                                         "checkme = f1()\n" << "int";
 
-    QTest::newRow("top_level_vs_class_member") << "var = 3\n"
-                                                  "class myclass:\n"
-                                                  "  def __init__(self): self.var = \"str\"\n"
-                                                  "  def f1(): return var\n"
-                                                  "checkme = myclass.f1()" << "int";
+    QTest::newRow("top_level_vs_class_attr") <<
+        "var = 3\n"
+        "class MyClass:\n"
+        "    var = 'str'\n"
+        "    def f1(): return var\n"
+        "checkme = MyClass.f1()" << "int";
+    QTest::newRow("top_level_vs_instance_attr") <<
+        "var = 3\n"
+        "class MyClass:\n"
+        "    def __init__(self): self.var = 'str'\n"
+        "    def f1(): return var\n"
+        "checkme = MyClass.f1()" << "int";
+    QTest::newRow("intermediate_vs_class/instance_attrs") <<
+        "def func():\n"
+        "    aa, bb = 3, 4\n"
+        "    class Foo:\n"
+        "        aa = 'a'\n"
+        "        def __init__(self):\n"
+        "            self.bb = 'b'\n"
+        "        def foo(self):\n"
+        "            return aa, bb\n"
+        "    return Foo().foo()\n"
+        "checkme = func()" << "tuple of (int, int)";
+    QTest::newRow("top_level_vs_nested_class_attrs") <<
+        "aaa = 'foo'\n"
+        "bbb = 'bar'\n"
+        "class Foo:\n"
+        "    aaa = 1\n"
+        "    class Bar:\n"
+        "        bbb = 2\n"
+        "        def foo(self, ccc=aaa, ddd=bbb):\n" // Bar.bbb is visible here, Foo.aaa isn't.
+        "            return ccc, ddd\n"
+        "checkme = Foo().Bar().foo()\n" << "tuple of (str, int)";
+    QTest::newRow("top_level_vs_nested_instance_attrs") <<
+        "aaa = 'foo'\n"
+        "bbb = 'bar'\n"
+        "class Foo:\n"
+        "    def __init__(self): self.aaa = 1\n"
+        "    class Bar:\n"
+        "        def __init__(self): self.bbb = 1\n"
+        "        def foo(self, ccc=aaa, ddd=bbb):\n" // self.bbb is visible here, Foo().aaa isn't.
+        "            return ccc, ddd\n"
+        "checkme = Foo().Bar().foo()\n" << "tuple of (str, int)";
 }
 
 typedef QPair<Declaration*, int> pair;
@@ -1240,6 +1324,7 @@ void PyDUChainTest::testProblemCount()
     QVERIFY(ctx);
 
     DUChainReadLocker lock;
+    QEXPECT_FAIL("fstring_visit_inside", "Ranges are broken so we don't visit the expression", Continue);
     QCOMPARE(ctx->problems().size(), problemsCount);
 }
 
@@ -1262,6 +1347,9 @@ void PyDUChainTest::testProblemCount_data()
     QTest::newRow("misplaced_return_value") << "return 15" << 1;
     QTest::newRow("misplaced_return_class") << "class A:\n return 25" << 1;
     QTest::newRow("correct_return") << "def foo():\n return" << 0;
+    QTest::newRow("lambda_argument_outside") << "def bar():\n lambda foo: 3\n foo" << 1;
+    QTest::newRow("use_found_at_decl") << "foo = 3" << 0;
+    QTest::newRow("fstring_visit_inside") << "checkme = f'{name}'" << 1;
 }
 
 void PyDUChainTest::testImportDeclarations_data() {
@@ -1521,6 +1609,10 @@ void PyDUChainTest::testContainerTypes()
     }
     else {
         QVERIFY(decls.first()->abstractType());
+        QEXPECT_FAIL("dict_of_int_call", "returnContentEqualsContentOf isn't suitable", Continue);
+        QEXPECT_FAIL("dict_from_tuples", "returnContentEqualsContentOf isn't suitable", Continue);
+        QEXPECT_FAIL("comprehension_shadowing_ms", "Nothing is foolproof to a sufficiently capable fool", Continue);
+        QEXPECT_FAIL("comprehension_shadowing_nest2", "See above", Continue);
         QCOMPARE(decls.first()->abstractType()->toString(), contenttype);
     }
 }
@@ -1534,16 +1626,26 @@ void PyDUChainTest::testContainerTypes_data()
     QTest::newRow("list_of_int") << "checkme = [1, 2, 3]" << "int" << false;
     QTest::newRow("list_from_unpacked") << "foo = [1.3]\ncheckme = [1, *foo, 3]" << "unsure (int, float)" << false;
     QTest::newRow("list_of_int_call") << "checkme = list([1, 2, 3])" << "int" << false;
+    QTest::newRow("list_from_tuple") << "checkme = list((1, 2, 3))" << "int" << false;
+    QTest::newRow("list_from_dict") << "checkme = list({'a':1, 'b':2})" << "str" << false; // Gets key type!
+    QTest::newRow("list_from_custom_iter") <<
+        "class MyClass:\n"
+        "    def __iter__(self): return self\n"
+        "    def __next__(self): return 3.1417\n"
+        "checkme = list(MyClass())" << "float" << false;
     QTest::newRow("generator") << "checkme = [i for i in [1, 2, 3]]" << "int" << false;
     QTest::newRow("list_access") << "list = [1, 2, 3]\ncheckme = list[0]" << "int" << true;
     QTest::newRow("set_of_int") << "checkme = {1, 2, 3}" << "int" << false;
     QTest::newRow("set_of_int_call") << "checkme = set({1, 2, 3})" << "int" << false;
+    QTest::newRow("set_from_tuple") << "checkme = set((1, 2, 3))" << "int" << false;
     QTest::newRow("set_generator") << "checkme = {i for i in [1, 2, 3]}" << "int" << false;
+    QTest::newRow("dict_of_str_int") << "checkme = {'a':1, 'b':2, 'c':3}" << "dict of str : int" << true;
     QTest::newRow("frozenset_of_int_call") << "checkme = frozenset({1, 2, 3})" << "int" << false;
     QTest::newRow("dict_of_int") << "checkme = {a:1, b:2, c:3}" << "int" << false;
-    QTest::newRow("dict_of_int_call") << "checkme = dict({a:1, b:2, c:3})" << "int" << false;
+    QTest::newRow("dict_of_int_call") << "checkme = dict({'a':1, 'b':2, 'c':3})" << "dict of str : int" << true;
+    QTest::newRow("dict_from_tuples") << "checkme = dict([('a', 1), ('b', 2)])" << "dict of str : int" << true;
     QTest::newRow("dict_generator") << "checkme = {\"Foo\":i for i in [1, 2, 3]}" << "int" << false;
-    QTest::newRow("dict_access") << "list = {a:1, b:2, c:3}\ncheckme = list[0]" << "int" << true;
+    QTest::newRow("dict_access") << "list = {'a':1, 'b':2, 'c':3}\ncheckme = list[0]" << "int" << true;
 #if PYTHON_VERSION >= QT_VERSION_CHECK(3, 5, 0)
     QTest::newRow("set_from_unpacked") << "foo = [1.3]\ncheckme = {1, *foo, 3}" << "unsure (int, float)" << false;
     QTest::newRow("dict_from_unpacked") << "checkme = {**{'a': 1}}" << "dict of str : int" << true;
@@ -1553,9 +1655,15 @@ void PyDUChainTest::testContainerTypes_data()
     QTest::newRow("generator_attribute") << "checkme = [item.capitalize() for item in ['foobar']]" << "str" << false;
     QTest::newRow("cannot_change_type") << "checkme = [\"Foo\", \"Bar\"]" << "str" << false;
     QTest::newRow("cannot_change_type2") << "[1, 2, 3].append(5)\ncheckme = [\"Foo\", \"Bar\"]" << "str" << false;
-    
+
     QTest::newRow("list_append") << "d = []\nd.append(3)\ncheckme = d[0]" << "int" << true;
     QTest::newRow("list_extend") << "d = []; q = [int()]\nd.extend(q)\ncheckme = d[0]" << "int" << true;
+    QTest::newRow("list_extend_with_tuple") << "d = []; q = (1, 2)\nd.extend(q)\ncheckme = d[0]" << "int" << true;
+    QTest::newRow("list_extend_with_custom_iter") <<
+        "class MyClass:\n"
+        "    def __iter__(self): return self\n"
+        "    def __next__(self): return 3.1417\n"
+        "checkme = []\ncheckme.extend(MyClass())" << "float" << false;
 
     QTest::newRow("for_loop") << "d = [3]\nfor item in d:\n checkme = item" << "int" << true;
     QTest::newRow("for_loop_unsure") << "d = [3, \"foo\"]\nfor item in d:\n checkme = item" << "unsure (int, str)" << true;
@@ -1568,6 +1676,14 @@ void PyDUChainTest::testContainerTypes_data()
                                             "sorted_list = sorted(users.items(), key=lambda kv: (-kv[1], kv[0]))\n"
                                             "checkme = [k for r,(k,v) in enumerate(sorted_list, 1)]" << "list of str" << true;
     QTest::newRow("comprehension_multiline") << "checkme = [a for\n a in \n (1, 2)]" << "list of int" << true;
+    QTest::newRow("comprehension_multistage") << "nested = (1, 2), (3, 4)\n"
+                                                 "checkme =  [foo for bar in nested for foo in bar]" << "list of int" << true;
+    QTest::newRow("comprehension_shadowing_ms") << "nested = (1, 2), (3, 4)\n"
+                                                   "checkme =  [foo for foo in nested for foo in foo]" << "list of int" << true;
+    QTest::newRow("comprehension_shadowing_nest1") << "nested = (1, 2), (3, 4)\n"
+                                                      "checkme = [foo for foo in [foo for foo in nested]]" << "list of tuple of (int, int)" << true;
+    QTest::newRow("comprehension_shadowing_nest2") << "nested = (1, 2), (3, 4)\n"
+                                                      "checkme = [[foo for foo in foo] for foo in nested]" << "list of list of int" << true;
     // From https://bugs.kde.org/show_bug.cgi?id=359912
     QTest::newRow("subscript_multi") <<
         "class Middle:\n def __getitem__(self, key):\n  return str()\n"
